/*
 * kb2mb2 --
 * redirect keypress/keyrelease event to mousebutton2
 *
 * Copyright (c) 1997  Tatsumi Hosokawa <hosokawa@jp.FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: kb2mb2.c,v 1.4 1997/12/24 06:26:49 hosokawa Exp $
 */
/*
 * Tested on XFree86 3.3.1 (w/ FreeBSD 2.2.5-RELEASE)
 * I think it'll work on Accel-X and Linux, etc.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

static int fmode = 0;
static int verbose = 0;
static int nondef = 0;
static KeyCode keycode = AnyKey;

#define DEFAULT_KEYCODE		0x74	/* Right Windoze key :-) */

extern char *optarg;
extern int optind;

void
usage(char *name)
{
    fprintf(stderr, "%s [-h|-v|-f|-c keycode]\n", name);
    fprintf(stderr, "  yet another \"Emulate3Button\" utility for XFree86.\n");
    fprintf(stderr, "Default:\n");
    fprintf(stderr, "  redirect right Windoze button to the second mouse button.\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -h: print this message.\n");
    fprintf(stderr, "  -v: verbose.\n");
    fprintf(stderr, "  -f: use the key pressed first as the second mouse button.\n");
    fprintf(stderr, "  -c keycode: use the key whose keycode is given value as the second \n");
    fprintf(stderr, "              mouse button.\n");
    exit(1);
}

int 
main(int argc, char *argv[])
{
    Display *d;
    Window w;
    XEvent ev;
    KeySym input_keysym;
    KeyCode input_keycode;
    long kc;
    int ch;
    char *dname = NULL;

    while ((ch = getopt(argc, argv, "hfvc:")) != EOF) {
	switch (ch) {
	  case 'f':
	    if (nondef) {
		fprintf(stderr, "Can't use -c and -f option simultaneously.\n");
		exit(1);
	    }
	    fmode = 1;
	    break;
	  case 'v':
	    verbose = 1;
	    break;
	  case 'h':
	    usage(argv[0]);
	    exit(0);
	    break;
	  case 'c':
	    if (fmode) {
		fprintf(stderr, "Can't use -c and -f option simultaneously.\n");
		exit(1);
	    }
	    kc = strtol(optarg, NULL, 0);
	    if (kc == LONG_MAX || kc == LONG_MIN || kc == 0) {
		usage(argv[0]);
		exit(1);
	    }
	    if (verbose) {
		fprintf(stderr, "Keycode = 0x%x\n", kc);
	    }
	    keycode = kc;
	    nondef = 1;
	    break;
	  default:
	    usage(argv[0]);
	    exit(1);
	}
    }

    argc -= optind;
    argv += optind;

    if (!(d = XOpenDisplay(dname))) {
	fprintf(stderr, "Can't Open Display.\n");
	exit(1);
    }

    w = RootWindow(d, DefaultScreen(d));
    XTestGrabControl(d, True);
    XSelectInput(d, w, KeyPressMask);

    if (!fmode) {
	if (!nondef) {
	    keycode = DEFAULT_KEYCODE;
	}
	if (verbose)
	  fprintf(stderr, "keycode = 0x%x", keycode);
    }
    else {
	fprintf(stderr, "Press the key you want to use as MouseButton2.\n", keycode);
	XGrabKeyboard(d, w, True, GrabModeAsync, GrabModeAsync, CurrentTime);
	while (1) {
	    XNextEvent(d, &ev);
	    switch (ev.type) {
	      case KeyPress:
		input_keycode = ev.xkey.keycode;
		input_keysym = XKeycodeToKeysym(d, input_keycode, 0);
		fprintf(stderr, "Okay, use keycode 0x%x as MouseButton2!\n", input_keycode);
		keycode = input_keycode;
		goto keypressed;
	    }
	}
keypressed:
	XUngrabKeyboard(d, CurrentTime);
	XFlush(d);
    } 
    
    XGrabKey(d, keycode, AnyModifier, w, True, GrabModeAsync, GrabModeAsync);

    while (1) {
	XNextEvent(d, &ev);
	switch (ev.type) {
	  case KeyPress:
	    input_keycode = ev.xkey.keycode;
	    input_keysym = XKeycodeToKeysym(d, input_keycode, 0);
	    XAutoRepeatOff(d);
	    if (verbose) {
		fprintf(stderr, "Pressed Key %x\n", input_keysym);
	    }
	    XTestFakeButtonEvent(d, Button2, True, CurrentTime);
	    break;
	  case KeyRelease:
	    input_keycode = ev.xkey.keycode;
	    input_keysym = XKeycodeToKeysym(d, input_keycode, 0);
	    XAutoRepeatOn(d);
	    if (verbose) {
		fprintf(stderr, "Released Key %x\n", input_keysym);
	    }
	    XTestFakeButtonEvent(d, Button2, False, CurrentTime);
	    break;
	}
	XFlush(d);
    }
}
